#!usr/bin/env python
# -*- coding: utf-8 -*-
import logging
import os
import argparse
import ast
from typing import Optional
import yaml
import click
from deprecated import deprecated

from GANDLF.entrypoints import append_copyright_to_help
from GANDLF.utils import writeTrainingCSV

from GANDLF.cli import copyrightMessage
from GANDLF.utils import logger_setup


def _construct_csv(
    input_dir: str,
    channels_id: str,
    label_id: Optional[str],
    output_file: str,
    relativize_paths_to_output: bool,
):
    input_dir = os.path.normpath(input_dir)
    output_file = os.path.normpath(output_file)

    # Do some special handling for if users pass a yml file for channel/label IDs
    # This is used for MLCube functionality because MLCube does not support plain string inputs.
    if channels_id.endswith(".yml") or channels_id.endswith(".yaml"):
        if os.path.isfile(channels_id):
            with open(channels_id, "r") as f:
                content = yaml.safe_load(f)
                channels_id = content["channels"]
                if isinstance(channels_id, list):
                    channels_id = ",".join(channels_id)

                # TODO: raise a warning if label param is both passed as arg and defined in file
                if "label" in content:
                    label_id = content["label"]
                    if isinstance(label_id, list):  # TODO: it can be really a list?
                        label_id = ",".join(label_id)

    logging.debug(f"{input_dir=}")
    logging.debug(f"{channels_id=}")
    logging.debug(f"{label_id=}")
    logging.debug(f"{output_file=}")
    logging.debug(f"{relativize_paths_to_output=}")

    writeTrainingCSV(
        input_dir, channels_id, label_id, output_file, relativize_paths_to_output
    )


@click.command()
@click.option(
    "--input-dir",
    "-i",
    required=True,
    type=click.Path(exists=True, file_okay=False, dir_okay=True),
    help="Input data directory which contains images in specified format",
)
@click.option(
    "--channels-id",
    "-c",
    required=True,
    help="Channels/modalities identifier string to check for in all files in 'input_dir'; for example: "
    "--channels-id _t1.nii.gz,_t2.nii.gz. May be a YAML file with `channels` list of suffixes",
    type=str,
)
@click.option(
    "--label-id",
    "-l",
    type=str,
    help="Label/mask identifier string to check for in all files in 'input_dir'; for example: "
    "--label-id _seg.nii.gz. Param value is ignored in `label` is defined in channels YAML file",
)
@click.option(
    "--output-file",
    "-o",
    required=True,
    type=click.Path(file_okay=True, dir_okay=False),
    help="Output CSV file",
)
@click.option(
    "--relativize-paths",
    "-r",
    is_flag=True,
    help="If True, paths in the output data CSV will always be relative to the location"
    " of the output data CSV itself.",
)
@click.option(
    "--log-file",
    type=click.Path(),
    default=None,
    help="Output file which will contain the logs.",
)
@append_copyright_to_help
def new_way(
    input_dir: str,
    channels_id: str,
    label_id: Optional[str],
    output_file: str,
    relativize_paths: bool,
    log_file: str,
):
    """Generate training/inference CSV from data directory."""

    logger_setup(log_file)
    _construct_csv(
        input_dir=input_dir,
        channels_id=channels_id,
        label_id=label_id,
        output_file=output_file,
        relativize_paths_to_output=relativize_paths,
    )


@deprecated(
    "This is a deprecated way of running GanDLF. Please, use `gandlf construct-csv` cli command "
    + "instead of `gandlf_constructCSV`. Note that in new CLI tool some params were renamed:\n"
    + "  --inputDir to --input-dir\n"
    + "  --channelsID to --channels-id\n"
    + "  --labelID to --label-id\n"
    + "  --outputFile to --output-file\n"
    + "  --relativizePaths to --relativize-paths and converted to flag, i.e. no value required\n"
    + "`gandlf_constructCSV` script would be deprecated soon."
)
def old_way():
    logger_setup()
    parser = argparse.ArgumentParser(
        prog="GANDLF_ConstructCSV",
        formatter_class=argparse.RawTextHelpFormatter,
        description="Generate training/inference CSV from data directory.\n\n"
        + copyrightMessage,
    )
    parser.add_argument(
        "-i",
        "--inputDir",
        metavar="",
        type=str,
        help="Input data directory which contains images in specified format",
    )
    parser.add_argument(
        "-c",
        "--channelsID",
        metavar="",
        type=str,
        help="Channels/modalities identifier string to check for in all files in 'input_dir'; for example: --channelsID _t1.nii.gz,_t2.nii.gz",
    )
    parser.add_argument(
        "-l",
        "--labelID",
        default=None,
        type=str,
        help="Label/mask identifier string to check for in all files in 'input_dir'; for example: --labelID _seg.nii.gz",
    )
    parser.add_argument(
        "-o", "--outputFile", metavar="", type=str, help="Output CSV file"
    )
    parser.add_argument(
        "-r",
        "--relativizePaths",
        metavar="",
        type=ast.literal_eval,
        default=False,
        help="If True, paths in the output data CSV will always be relative to the location of the output data CSV itself.",
    )

    args = parser.parse_args()

    # check for required parameters - this is needed here to keep the cli clean
    for param_name in ["inputDir", "channelsID", "outputFile"]:
        param_none_check = getattr(args, param_name)
        assert param_none_check is not None, f"Missing required parameter: {param_name}"

    _construct_csv(
        input_dir=args.inputDir,
        channels_id=args.channelsID,
        label_id=args.labelID,
        output_file=args.outputFile,
        relativize_paths_to_output=args.relativizePaths,
    )


# main function
if __name__ == "__main__":
    old_way()



import sys
import torch
from .loss_interface import AbstractSegmentationLoss, AbstractLossFunction


class MulticlassDiceLoss(AbstractSegmentationLoss):
    """
    This class computes the Dice loss between two tensors.
    """

    def _single_class_loss_calculator(
        self, prediction: torch.Tensor, target: torch.Tensor
    ) -> torch.Tensor:
        """
        Compute Dice score for a single class.

        Args:
            prediction (torch.Tensor): Network's predicted segmentation mask
            target (torch.Tensor): Target segmentation mask

        Returns:
            torch.Tensor: The computed dice score.
        """
        predicted_flat = prediction.flatten()
        label_flat = target.flatten()
        intersection = (predicted_flat * label_flat).sum()

        dice_score = (2.0 * intersection + sys.float_info.min) / (
            predicted_flat.sum() + label_flat.sum() + sys.float_info.min
        )

        return dice_score


class MulticlassDiceLogLoss(MulticlassDiceLoss):
    def _optional_loss_operations(self, loss):
        return -torch.log(
            loss + torch.finfo(torch.float32).eps
        )  # epsilon for numerical stability


class MulticlassMCCLoss(AbstractSegmentationLoss):
    """
    This class computes the Matthews Correlation Coefficient (MCC) loss between two tensors.
    """

    def _single_class_loss_calculator(
        self, prediction: torch.Tensor, target: torch.Tensor
    ) -> torch.Tensor:
        """
        Compute MCC score for a single class.

        Args:
            prediction (torch.Tensor): Network's predicted segmentation mask
            target (torch.Tensor): Target segmentation mask

        Returns:
            torch.Tensor: The computed MCC score.
        """
        tp = torch.sum(torch.mul(prediction, target))
        tn = torch.sum(torch.mul((1 - prediction), (1 - target)))
        fp = torch.sum(torch.mul(prediction, (1 - target)))
        fn = torch.sum(torch.mul((1 - prediction), target))

        numerator = torch.mul(tp, tn) - torch.mul(fp, fn)
        # Adding epsilon to the denominator to avoid divide-by-zero errors.
        denominator = (
            torch.sqrt(
                torch.add(tp, 1, fp)
                * torch.add(tp, 1, fn)
                * torch.add(tn, 1, fp)
                * torch.add(tn, 1, fn)
            )
            + torch.finfo(torch.float32).eps
        )

        return torch.div(numerator.sum(), denominator.sum())


class MulticlassMCLLogLoss(MulticlassMCCLoss):
    def _optional_loss_operations(self, loss):
        return -torch.log(
            loss + torch.finfo(torch.float32).eps
        )  # epsilon for numerical stability


class MulticlassTverskyLoss(AbstractSegmentationLoss):
    """
    This class computes the Tversky loss between two tensors.
    """

    def __init__(self, params: dict):
        super().__init__(params)
        loss_params = params["loss_function"]
        self.alpha = 0.5
        self.beta = 0.5
        if isinstance(loss_params, dict):
            self.alpha = loss_params.get("alpha", self.alpha)
            self.beta = loss_params.get("beta", self.beta)

    def _single_class_loss_calculator(
        self, prediction: torch.Tensor, target: torch.Tensor
    ) -> torch.Tensor:
        """
        Compute Tversky score for a single class.

        Args:
            prediction (torch.Tensor): Network's predicted segmentation mask
            target (torch.Tensor): Target segmentation mask

        Returns:
            torch.Tensor: The computed Tversky score.
        """
        predicted_flat = prediction.contiguous().view(-1)
        target_flat = target.contiguous().view(-1)

        true_positives = (predicted_flat * target_flat).sum()
        false_positives = ((1 - target_flat) * predicted_flat).sum()
        false_negatives = (target_flat * (1 - predicted_flat)).sum()

        numerator = true_positives
        denominator = (
            true_positives + self.alpha * false_positives + self.beta * false_negatives
        )
        loss = (numerator + sys.float_info.min) / (denominator + sys.float_info.min)

        return loss


class MulticlassFocalLoss(AbstractSegmentationLoss):
    """
    This class computes the Focal loss between two tensors.
    """

    def __init__(self, params: dict):
        super().__init__(params)

        self.ce_loss_helper = torch.nn.CrossEntropyLoss(reduction="none")
        loss_params = params["loss_function"]
        self.alpha = 1.0
        self.gamma = 2.0
        self.output_aggregation = "sum"
        if isinstance(loss_params, dict):
            self.alpha = loss_params.get("alpha", self.alpha)
            self.gamma = loss_params.get("gamma", self.gamma)
            self.output_aggregation = loss_params.get(
                "size_average",
                self.output_aggregation,  # naming mismatch of key due to keeping API consistent with config format
            )
        assert self.output_aggregation in [
            "sum",
            "mean",
        ], f"Invalid output aggregation method defined for Foal Loss: {self.output_aggregation}. Valid options are ['sum', 'mean']"

    def _single_class_loss_calculator(
        self, prediction: torch.Tensor, target: torch.Tensor
    ) -> torch.Tensor:
        """
        Compute focal loss for a single class. It is based on the following formulas:
            FocalLoss(p_t) = -alpha_t * (1 - p_t)^gamma * log(p_t)
            CrossEntropy(pred, target) = -log(pred) if target = 1 else -log(1 - pred)
            CrossEntropy(p_t) = CrossEntropy(pred, target) = -log(p_t)
            p_t = p if target = 1 else 1 - p
        """
        ce_loss = self.ce_loss_helper(prediction, target)
        p_t = torch.exp(-ce_loss)
        loss = -self.alpha * (1 - p_t) ** self.gamma * ce_loss
        return loss.sum() if self.output_aggregation == "sum" else loss.mean()

    def _compute_single_class_loss(
        self, prediction: torch.Tensor, target: torch.Tensor, class_idx: int
    ) -> torch.Tensor:
        """Compute loss for a single class."""
        loss_value = self._single_class_loss_calculator(
            prediction[:, class_idx, ...], target[:, class_idx, ...]
        )
        return loss_value  # no need to subtract from 1 in this case, hence the override


class KullbackLeiblerDivergence(AbstractLossFunction):
    def forward(self, mu: torch.Tensor, logvar: torch.Tensor) -> torch.Tensor:
        """
        Calculates the Kullback-Leibler divergence between two Gaussian distributions.

        Args:
            mu (torch.Tensor): The mean of the first Gaussian distribution.
            logvar (torch.Tensor): The logarithm of the variance of the first Gaussian distribution.

        Returns:
            torch.Tensor: The computed Kullback-Leibler divergence
        """
        loss = -0.5 * torch.sum(1 + logvar - mu.pow(2) - logvar.exp(), dim=-1)
        return loss.mean()



"""
All the segmentation metrics are to be called from here
"""

from typing import List, Optional, Tuple, Union
import sys
import torch
import numpy as np
from GANDLF.losses.segmentation import dice
from scipy.ndimage import _ni_support
from scipy.ndimage.morphology import (
    distance_transform_edt,
    binary_erosion,
    generate_binary_structure,
)


def _convert_tensor_to_int_label_array(input_tensor: torch.Tensor) -> np.ndarray:
    """
    This function converts a tensor of labels to a numpy array of labels.

    Args:
        input_tensor (torch.Tensor): Input data containing objects. Can be any type but will be converted into binary: background where 0, object everywhere else.

    Returns:
        numpy.ndarray: The numpy array of labels.
    """
    result_array = input_tensor.detach().cpu().numpy()
    if result_array.shape[-1] == 1:
        result_array = result_array.squeeze(-1)
    # ensure that we are dealing with a binary array
    result_array[result_array < 0.5] = 0
    result_array[result_array >= 0.5] = 1
    return result_array.astype(np.int64)


def multi_class_dice(
    prediction: torch.Tensor,
    target: torch.Tensor,
    params: dict,
    per_label: Optional[bool] = False,
) -> Union[torch.Tensor, List[float]]:
    """
    This function computes a multi-class dice.

    Args:
        prediction (torch.Tensor): The input prediction containing objects. Can be any type but will be converted into binary: background where 0, object everywhere else.
        target (torch.Tensor): The input ground truth containing objects. Can be any type but will be converted into binary: background where 0, object everywhere else.
        params (dict): The parameter dictionary containing training and data information.
        per_label (Optional[bool], optional): Whether to return per-label scores. Defaults to False.

    Returns:
        Union[torch.Tensor, List[float]]: The multi-class dice score or the list of per-label dice scores.
    """
    total_dice = 0
    avg_counter = 0
    per_label_dice = []
    for i in range(0, params["model"]["num_classes"]):
        # this check should only happen during validation
        if i != params["model"]["ignore_label_validation"]:
            current_dice = dice(prediction[:, i, ...], target[:, i, ...])
            total_dice += current_dice
            per_label_dice.append(current_dice.item())
            avg_counter += 1
        # currentDiceLoss = 1 - currentDice # subtract from 1 because this is a loss
    total_dice /= avg_counter

    if per_label:
        return torch.tensor(per_label_dice)
    else:
        return total_dice


def multi_class_dice_per_label(
    prediction: torch.Tensor, target: torch.Tensor, params: dict
) -> list:
    """
    This function computes a multi-class dice.

    Args:
        prediction (torch.Tensor): Input data containing objects. Can be any type but will be converted into binary: background where 0, object everywhere else.
        target (torch.Tensor): Input data containing objects. Can be any type but will be converted into binary: background where 0, object everywhere else.
        params (dict): The parameter dictionary containing training and data information.

    Returns:
        list: The list of per-label dice scores.
    """
    return multi_class_dice(prediction, target, params, per_label=True)


def __surface_distances(
    prediction: torch.Tensor,
    target: torch.Tensor,
    voxel_spacing: Optional[Tuple[float]] = None,
    connectivity: Optional[int] = 1,
) -> float:
    """
    The distances between the surface voxel of binary objects in result and their
    nearest partner surface voxel of a binary object in reference. Adapted from https://github.com/loli/medpy/blob/39131b94f0ab5328ab14a874229320efc2f74d98/medpy/metric/binary.py#L1195.

    Args:
        prediction (torch.Tensor): Input prediction containing objects. Can be any type but will be converted into binary: background where 0, object everywhere else.
        target (torch.Tensor): Input ground truth containing objects. Can be any type but will be converted into binary: background where 0, object everywhere else.
        voxel_spacing (Optional[Tuple[float]], optional): The voxel spacing. Defaults to None.
        connectivity (Optional[int], optional): The voxel connectivity. Defaults to 1.

    Returns:
        float: The symmetric Hausdorff Distance between the object(s) in ```result``` and the object(s) in ```reference```. The distance unit is the same as for the spacing of elements along each dimension, which is usually given in mm.
    """
    result = np.atleast_1d(prediction.astype(bool))
    reference = np.atleast_1d(target.astype(bool))
    if voxel_spacing is not None:
        voxel_spacing = _ni_support._normalize_sequence(voxel_spacing, result.ndim)
        voxel_spacing = np.asarray(voxel_spacing, dtype=np.float64)
        if not voxel_spacing.flags.contiguous:
            voxel_spacing = voxel_spacing.copy()

    # binary structure
    footprint = generate_binary_structure(result.ndim, connectivity)

    # test for emptiness
    if 0 == np.count_nonzero(result):
        return 0
    if 0 == np.count_nonzero(reference):
        return 0

    # extract only 1-pixel border line of objects
    result_border = result ^ binary_erosion(result, structure=footprint, iterations=1)
    reference_border = reference ^ binary_erosion(
        reference, structure=footprint, iterations=1
    )

    # compute average surface distance
    # Note: scipys distance transform is calculated only inside the borders of the
    #       foreground objects, therefore the input has to be reversed
    dt = distance_transform_edt(~reference_border, sampling=voxel_spacing)
    sds = dt[result_border]

    return sds


def _nsd_base(a_to_b: np.ndarray, b_to_a: np.ndarray, threshold: float) -> float:
    """
    This implementation differs from the official surface dice implementation! These two are not comparable!!!!!
    The normalized surface dice is symmetric, so it should not matter whether a or b is the reference image
    This implementation natively supports 2D and 3D images.

    Args:
        a_to_b (np.ndarray): Surface distances from a to b
        b_to_a (np.ndarray): Surface distances from b to a
        threshold (float): distances below this threshold will be counted as true positives. Threshold is in mm, not voxels!

    Returns:
        float: the normalized surface dice between a and b
    """
    if isinstance(a_to_b, int):
        return 0
    if isinstance(b_to_a, int):
        return 0
    numel_a = len(a_to_b)
    numel_b = len(b_to_a)
    tp_a = np.sum(a_to_b <= threshold) / numel_a
    tp_b = np.sum(b_to_a <= threshold) / numel_b
    fp = np.sum(a_to_b > threshold) / numel_a
    fn = np.sum(b_to_a > threshold) / numel_b
    dc = (tp_a + tp_b) / (tp_a + tp_b + fp + fn + sys.float_info.min)
    return dc


def _calculator_jaccard(
    prediction: torch.Tensor,
    target: torch.Tensor,
    params: dict,
    per_label: Optional[bool] = False,
) -> torch.Tensor:
    """
    This function returns sensitivity and specificity.

    Args:
        prediction (torch.Tensor): Input prediction containing objects. Can be any type but will be converted into binary: background where 0, object everywhere else.
        target (torch.Tensor): Input ground truth containing objects. Can be any type but will be converted into binary: binary: background where 0, object everywhere else.
        params (dict): The parameter dictionary containing training and data information.
        per_label (Optional[bool], optional): Whether to return per-label scores. Defaults to False.

    Returns:
        float: The Jaccard score between the object(s) in ```inp``` and the object(s) in ```target```.
    """
    result_array = _convert_tensor_to_int_label_array(prediction)
    target_array = _convert_tensor_to_int_label_array(target)

    jaccard, avg_counter = 0, 0
    jaccard_per_label = []
    for i in range(0, params["model"]["num_classes"]):
        # this check should only happen during validation
        if i != params["model"]["ignore_label_validation"]:
            dice_score = dice(result_array[:, i, ...], target_array[:, i, ...])
            # https://en.wikipedia.org/wiki/S%C3%B8rensen%E2%80%93Dice_coefficient#Difference_from_Jaccard
            j_score = dice_score / (2 - dice_score)
            jaccard += j_score
            if per_label:
                jaccard_per_label.append(jaccard)
            avg_counter += 1

    if per_label:
        return torch.tensor(jaccard_per_label)
    else:
        return torch.tensor(jaccard / avg_counter)


def _calculator_sensitivity_specificity(
    prediction: torch.Tensor,
    target: torch.Tensor,
    params: dict,
    per_label: Optional[bool] = False,
) -> torch.Tensor:
    """
    This function returns sensitivity and specificity.

    Args:
        prediction (torch.Tensor): Input prediction containing objects. Can be any type but will be converted into binary: background where 0, object everywhere else.
        target (torch.Tensor): Input ground truth containing objects. Can be any type but will be converted into binary: binary: background where 0, object everywhere else.
        params (dict): The parameter dictionary containing training and data information.
        per_label (Optional[bool], optional): Whether to return per-label scores. Defaults to False.

    Returns:
        float, float: The sensitivity and specificity between the object(s) in ```inp``` and the object(s) in ```target```.
    """
    # inMask is mask of input array equal to a certain tissue (ie. all one's in tumor core)
    # Ref mask is mask of certain tissue in ground truth (ie. all one's in reference core )
    # overlap is mask where the two equal each other
    # They are of the total number of voxels of the ground truth brain mask

    def get_sensitivity_and_specificity(result_array, target_array):
        iC = np.sum(result_array)
        rC = np.sum(target_array)

        overlap = np.where((result_array == target_array), 1, 0)

        # Where they agree are both equal to that value
        TP = overlap[result_array == 1].sum()
        FP = iC - TP
        FN = rC - TP
        TN = np.count_nonzero((result_array != 1) & (target_array != 1))

        Sens = 1.0 * TP / (TP + FN + sys.float_info.min)
        Spec = 1.0 * TN / (TN + FP + sys.float_info.min)

        # Make Changes if both input and reference are 0 for the tissue type
        if (iC == 0) and (rC == 0):
            Sens = 1.0

        return Sens, Spec

    result_array = _convert_tensor_to_int_label_array(prediction)
    target_array = _convert_tensor_to_int_label_array(target)

    sensitivity, specificity, avg_counter = 0, 0, 0
    sensitivity_per_label, specificity_per_label = [], []
    for b in range(0, result_array.shape[0]):
        for i in range(0, params["model"]["num_classes"]):
            if i != params["model"]["ignore_label_validation"]:
                s, p = get_sensitivity_and_specificity(
                    result_array[b, i, ...], target_array[b, i, ...]
                )
                sensitivity += s
                specificity += p
                if per_label:
                    sensitivity_per_label.append(s)
                    specificity_per_label.append(p)
                avg_counter += 1

    if per_label:
        return torch.tensor(sensitivity_per_label), torch.tensor(specificity_per_label)
    else:
        return (
            torch.tensor(sensitivity / avg_counter),
            torch.tensor(specificity / avg_counter),
        )


def _calculator_generic_all_surface_distances(
    prediction: torch.Tensor,
    target: torch.Tensor,
    params: dict,
    per_label: Optional[bool] = False,
) -> torch.Tensor:
    """
    This function returns hd100, hd95, and nsd.

    Args:
        prediction (torch.Tensor): Input prediction containing objects. Can be any type but will be converted into binary: background where 0, object everywhere else.
        target (torch.Tensor): Input ground truth containing objects. Can be any type but will be converted into binary: binary: background where 0, object everywhere else.
        params (dict): The parameter dictionary containing training and data information.
        per_label (Optional[bool], optional): Whether to return per-label scores. Defaults to False.


    Returns:
        float, float, float: The Normalized Surface Dice, 100th percentile Hausdorff Distance, and the 95th percentile Hausdorff Distance.
    """
    result_array = _convert_tensor_to_int_label_array(prediction)
    target_array = _convert_tensor_to_int_label_array(target)

    avg_counter = 0
    if per_label:
        return_hd100, return_hd95, return_nsd = [], [], []
    else:
        return_hd100, return_hd95, return_nsd = 0, 0, 0
    for b in range(0, result_array.shape[0]):
        for i in range(0, params["model"]["num_classes"]):
            if i != params["model"]["ignore_label_validation"]:
                hd1 = __surface_distances(
                    result_array[b, i, ...],
                    target_array[b, i, ...],
                    params["subject_spacing"][b],
                )
                hd2 = __surface_distances(
                    target_array[b, i, ...],
                    result_array[b, i, ...],
                    params["subject_spacing"][b],
                )
                threshold = max(min(params["subject_spacing"][0]), 1).item()
                temp_nsd = _nsd_base(hd1, hd2, threshold)
                temp_hd100 = np.percentile(np.hstack((hd1, hd2)), 100)
                temp_hd95 = np.percentile(np.hstack((hd1, hd2)), 95)
                if per_label:
                    return_nsd.append(temp_nsd)
                    return_hd100.append(temp_hd100)
                    return_hd95.append(temp_hd95)
                else:
                    return_nsd += temp_nsd
                    return_hd100 += temp_hd100
                    return_hd95 += temp_hd95
                    avg_counter += 1

    if per_label:
        return (
            torch.tensor(return_nsd),
            torch.tensor(return_hd100),
            torch.tensor(return_hd95),
        )
    else:
        return (
            torch.tensor(return_nsd / avg_counter),
            torch.tensor(return_hd100 / avg_counter),
            torch.tensor(return_hd95 / avg_counter),
        )


def _calculator_generic(
    prediction: torch.Tensor,
    target: torch.Tensor,
    params: dict,
    percentile: int = 95,
    surface_dice: Optional[bool] = False,
    per_label: Optional[bool] = False,
) -> Union[torch.Tensor, List[float]]:
    """
    Generic Surface Dice (SD)/Hausdorff (HD) Distance calculation from 2 tensors. Compared to the standard Hausdorff Distance, this metric is slightly more stable to small outliers and is commonly used in Biomedical Segmentation challenges.

    Args:
        prediction (torch.Tensor): Input prediction containing objects. Can be any type but will be converted into binary: background where 0, object everywhere else.
        target (torch.Tensor): Input ground truth containing objects. Can be any type but will be converted into binary: binary: background where 0, object everywhere else.
        params (dict): The parameter dictionary containing training and data information.
        percentile (int, optional): The percentile to calculate the Hausdorff Distance. Defaults to 95.
        surface_dice (Optional[bool], optional): Whether to return the surface dice. Defaults to False.
        per_label (Optional[bool], optional): Whether to return per-label scores. Defaults to False.


    Returns:
        Union[torch.Tensor, List[float]]: The Normalized Surface Dice, 100th percentile Hausdorff Distance, and the 95th percentile Hausdorff Distance, or the list of per-label scores for each metric.
    """
    _nsd, _hd100, _hd95 = _calculator_generic_all_surface_distances(
        prediction, target, params, per_label=per_label
    )
    if surface_dice:
        return _nsd
    elif percentile == 95:
        return _hd95
    else:
        return _hd100


def hd95(prediction: torch.Tensor, target: torch.Tensor, params: dict) -> torch.Tensor:
    """
    This function returns the 95th percentile Hausdorff Distance.

    Args:
        prediction (torch.Tensor): Input prediction containing objects. Can be any type but will be converted into binary: background where 0, object everywhere else.
        target (torch.Tensor): Input ground truth containing objects. Can be any type but will be converted into binary: binary: background where 0, object everywhere else.
        params (dict): The parameter dictionary containing training and data information.

    Returns:
        torch.Tensor: The 95th percentile Hausdorff Distance.
    """
    return _calculator_generic(prediction, target, params, percentile=95)


def hd95_per_label(
    prediction: torch.Tensor, target: torch.Tensor, params: dict
) -> List[float]:
    """
    This function returns the per-label 95th percentile Hausdorff Distance.

    Args:
        prediction (torch.Tensor): Input prediction containing objects. Can be any type but will be converted into binary: background where 0, object everywhere else.
        target (torch.Tensor): Input ground truth containing objects. Can be any type but will be converted into binary: binary: background where 0, object everywhere else.
        params (dict): The parameter dictionary containing training and data information.

    Returns:
        List[float]: The list of per-label 95th percentile Hausdorff Distances.
    """
    return _calculator_generic(
        prediction, target, params, percentile=95, per_label=True
    )


def hd100(prediction: torch.Tensor, target: torch.Tensor, params: dict) -> torch.Tensor:
    """
    This function returns the 100th percentile Hausdorff Distance.

    Args:
        prediction (torch.Tensor): Input prediction containing objects. Can be any type but will be converted into binary: background where 0, object everywhere else.
        target (torch.Tensor): Input ground truth containing objects. Can be any type but will be converted into binary: binary: background where 0, object everywhere else.
        params (dict): The parameter dictionary containing training and data information.

    Returns:
        torch.Tensor: The 100th percentile Hausdorff Distance.
    """
    return _calculator_generic(prediction, target, params, percentile=100)


def hd100_per_label(
    prediction: torch.Tensor, target: torch.Tensor, params: dict
) -> List[float]:
    """
    This function returns the per-label 100th percentile Hausdorff Distance.

    Args:
        prediction (torch.Tensor): Input prediction containing objects. Can be any type but will be converted into binary: background where 0, object everywhere else.
        target (torch.Tensor): Input ground truth containing objects. Can be any type but will be converted into binary: binary: background where 0, object everywhere else.
        params (dict): The parameter dictionary containing training and data information.

    Returns:
        List[float]: The list of per-label 100th percentile Hausdorff Distances.
    """
    return _calculator_generic(
        prediction, target, params, percentile=100, per_label=True
    )


def nsd(prediction: torch.Tensor, target: torch.Tensor, params: dict) -> torch.Tensor:
    """
    This function returns the Normalized Surface Dice.

    Args:
        prediction (torch.Tensor): Input prediction containing objects. Can be any type but will be converted into binary: background where 0, object everywhere else.
        target (torch.Tensor): Input ground truth containing objects. Can be any type but will be converted into binary: binary: background where 0, object everywhere else.
        params (dict): The parameter dictionary containing training and data information.

    Returns:
        torch.Tensor: The Normalized Surface Dice.
    """
    return _calculator_generic(
        prediction, target, params, percentile=100, surface_dice=True
    )


def nsd_per_label(
    prediction: torch.Tensor, target: torch.Tensor, params: dict
) -> List[float]:
    """
    This function returns the per-label Normalized Surface Dice.

    Args:
        prediction (torch.Tensor): Input prediction containing objects. Can be any type but will be converted into binary: background where 0, object everywhere else.
        target (torch.Tensor): Input ground truth containing objects. Can be any type but will be converted into binary: binary: background where 0, object everywhere else.
        params (dict): The parameter dictionary containing training and data information.

    Returns:
        List[float]: The list of per-label Normalized Surface Dice scores.
    """
    return _calculator_generic(
        prediction, target, params, percentile=100, per_label=True, surface_dice=True
    )


def sensitivity(
    prediction: torch.Tensor, target: torch.Tensor, params: dict
) -> torch.Tensor:
    """
    This function returns the sensitivity.

    Args:
        prediction (torch.Tensor): Input prediction containing objects. Can be any type but will be converted into binary: background where 0, object everywhere else.
        target (torch.Tensor): Input ground truth containing objects. Can be any type but will be converted into binary: binary: background where 0, object everywhere else.
        params (dict): The parameter dictionary containing training and data information.

    Returns:
        torch.Tensor: The sensitivity.
    """
    s, _ = _calculator_sensitivity_specificity(prediction, target, params)
    return s


def sensitivity_per_label(
    prediction: torch.Tensor, target: torch.Tensor, params: dict
) -> List[float]:
    """
    This function returns the per-label sensitivity.

    Args:
        prediction (torch.Tensor): Input prediction containing objects. Can be any type but will be converted into binary: background where 0, object everywhere else.
        target (torch.Tensor): Input ground truth containing objects. Can be any type but will be converted into binary: binary: background where 0, object everywhere else.
        params (dict): The parameter dictionary containing training and data information.

    Returns:
        List[float]: The list of per-label sensitivity scores.
    """
    s, _ = _calculator_sensitivity_specificity(
        prediction, target, params, per_label=True
    )
    return s


def specificity_segmentation(
    prediction: torch.Tensor, target: torch.Tensor, params: dict
) -> torch.Tensor:
    """
    This function returns the specificity.

    Args:
        prediction (torch.Tensor): Input prediction containing objects. Can be any type but will be converted into binary: background where 0, object everywhere else.
        target (torch.Tensor): Input ground truth containing objects. Can be any type but will be converted into binary: binary: background where 0, object everywhere else.
        params (dict): The parameter dictionary containing training and data information.

    Returns:
        torch.Tensor: The specificity.
    """
    _, p = _calculator_sensitivity_specificity(prediction, target, params)
    return p


def specificity_segmentation_per_label(
    prediction: torch.Tensor, target: torch.Tensor, params: dict
) -> List[float]:
    """
    This function returns the per-label specificity.

    Args:
        prediction (torch.Tensor): Input prediction containing objects. Can be any type but will be converted into binary: background where 0, object everywhere else.
        target (torch.Tensor): Input ground truth containing objects. Can be any type but will be converted into binary: binary: background where 0, object everywhere else.
        params (dict): The parameter dictionary containing training and data information.

    Returns:
        List[float]: The list of per-label specificity scores.
    """
    _, p = _calculator_sensitivity_specificity(
        prediction, target, params, per_label=True
    )
    return p


def jaccard(
    prediction: torch.Tensor, target: torch.Tensor, params: dict
) -> torch.Tensor:
    """
    This function returns the Jaccard score.

    Args:
        prediction (torch.Tensor): Input prediction containing objects. Can be any type but will be converted into binary: background where 0, object everywhere else.
        target (torch.Tensor): Input ground truth containing objects. Can be any type but will be converted into binary: binary: background where 0, object everywhere else.
        params (dict): The parameter dictionary containing training and data information.

    Returns:
        torch.Tensor: The Jaccard score.
    """
    j = _calculator_jaccard(prediction, target, params)
    return j


def jaccard_per_label(
    prediction: torch.Tensor, target: torch.Tensor, params: dict
) -> List[float]:
    """
    This function returns the per-label Jaccard score.

    Args:
        prediction (torch.Tensor): Input prediction containing objects. Can be any type but will be converted into binary: background where 0, object everywhere else.
        target (torch.Tensor): Input ground truth containing objects. Can be any type but will be converted into binary: binary: background where 0, object everywhere else.
        params (dict): The parameter dictionary containing training and data information.

    Returns:
        List[float]: The list of per-label Jaccard scores.
    """
    j = _calculator_jaccard(prediction, target, params, per_label=True)
    return j
